#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include "Castor_limits.h"
#include "dpns_api.h"
#include "dpm_api.h"
#include "rfio_api.h"
#include "serrno.h"
#include "voms.h"
extern int delay_signaling(void);
extern gid_t dpm_gid;
extern uid_t dpm_uid;
extern int enable_signaling(void);
extern char data_interface[];
extern int rfio_parse(char *, char **, char **);

extern gid_t Csec_gid;
extern uid_t Csec_uid;
extern gssapi_voms_ctx_t voms_ctx;

struct stage_hsm {	/* Should be identical to the one in rfio_HsmIf.c */
	char	*sfn;
	char	*r_token;
};
typedef struct stage_hsm stage_hsm_t;
extern int rfio_HsmIf_AddCnsFileDescriptor(int, int, stage_hsm_t *);
extern int rfio_HsmIf_SetCnsWrittenTo(int);

int dpm_handle_open(char *path, int flags, int mode, char *id)
{
	uid_t euid;
	char *filename;
	char *host;
	char *p;
	int rc;
	char sfn1[CA_MAXSFNLEN+1];
	struct stat64 st;
	char *turl;
	mode_t oldmask;

	if ((rc = rfio_parse (path, &host, &filename)) < 0)
		return (-1);
	euid = geteuid();

	if (rc) {	/* remote file */
		if (Csec_uid) {
			rfio_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				rfio_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		rc = rfio_open64 (path, flags, mode);
		seteuid (euid);
		return (rc);
	}

	if (host) {	/* DPM managed file */
		stage_hsm_t *hsmfile = NULL;
		hsmfile = (stage_hsm_t *)calloc (1, sizeof(stage_hsm_t));
		hsmfile->sfn = strdup (path);
		hsmfile->r_token = malloc (CA_MAXDPMTOKENLEN+1);
		if (Csec_uid) {
			dpm_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				dpm_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		turl = dpm_getturl (path, flags, (u_signed64) 0, hsmfile->r_token);
		seteuid (euid);
		if (turl) {
			path = turl + 7;
			p = strchr (path, '/');
			*p = ':';
			if (rfio_parse (path, &host, &filename)) {
				if (Csec_uid) {
					rfio_client_setAuthorizationId (Csec_uid,
					    Csec_gid, "GSI", id);
					if (voms_ctx.voname && voms_ctx.fqan)
						rfio_client_setVOMS_data (voms_ctx.voname,
							voms_ctx.fqan, voms_ctx.nbfqan);
				}
				seteuid (0);
				rc = rfio_open64 (path, flags, mode);
				seteuid (euid);
			} else {
				oldmask = umask (0);
				delay_signaling();
				seteuid (0);
				rc = open64 (filename, flags, 0660);
				if (flags & O_CREAT)
					chown (filename, dpm_uid, dpm_gid);
				seteuid (euid);
				enable_signaling();
				umask (oldmask);
			}
			free (turl);
			if (rc >= 0) {
				if (rfio_HsmIf_AddCnsFileDescriptor (rc, flags, hsmfile) < 0) {
					rfio_close (rc);
					serrno = SEINTERNAL;
					rc = -1;
				} else {
					if (flags & O_CREAT) rfio_HsmIf_SetCnsWrittenTo (rc);
				}
			}
		} else
			rc = -1;
		if (rc < 0) {
			if (hsmfile->sfn) free (hsmfile->sfn);
			if (hsmfile->r_token) free (hsmfile->r_token);
			free (hsmfile);
		}
		return (rc);
	}

	/* local file */

	delay_signaling();
	seteuid (0);
	rc = stat64 (filename, &st);
	seteuid (euid);
	enable_signaling();
	if (rc < 0) {
		char *parent;
		if (errno != ENOENT || (flags & O_CREAT) == 0)
			return (-1);
		if ((parent = strdup (filename)) == NULL)
			return (-1);
		if ((p = strrchr (parent, '/')) == NULL) {
			free (parent);
			errno = EINVAL;
			return (-1);
		}
		if (p != parent) {
			*p = '\0';
			p = parent;
		} else
			p = "/";
		delay_signaling();
		seteuid (0);
		rc = stat64 (p, &st);
		seteuid (euid);
		enable_signaling();
		free (parent);
		if (rc < 0)
			return (-1);
	}
	if (st.st_uid != dpm_uid &&
	    st.st_gid != dpm_gid) {	/* not a file managed by DPM */
		return (open64 (filename, flags, mode));
	}

	/* DPM managed file */

	strcpy (sfn1, data_interface);
	p = sfn1 + strlen (sfn1);
	*p++ = ':';
	strcpy (p, filename);
	if (Csec_uid) {
		Cns_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
		if (voms_ctx.voname && voms_ctx.fqan)
			Cns_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
					voms_ctx.nbfqan);
	}
	seteuid (0);
	rc = Cns_accessr (sfn1, (flags & (O_WRONLY|O_TRUNC|O_CREAT)) ? W_OK : R_OK);
	seteuid (euid);
	if (rc < 0)
		return (-1);
	oldmask = umask (0);
	delay_signaling();
	seteuid (0);
	rc = open64 (filename, flags, 0660);
	if (flags & O_CREAT)
		chown (filename, dpm_uid, dpm_gid);
	seteuid (euid);
	enable_signaling();
	umask (oldmask);
	return (rc);
}

int dpm_handle_stat(char *path, struct stat64 *st, char *id)
{
	uid_t euid;
	char *filename;
	char *host;
	char *p;
	int rc;
	char sfn1[CA_MAXSFNLEN+1];
	struct Cns_filestatg statbuf;
															     
	if ((rc = rfio_parse (path, &host, &filename)) < 0)
		return (-1);
	euid = geteuid();

	if (rc) {       /* remote file */
		if (Csec_uid) {
			rfio_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				rfio_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		rc = rfio_stat64 (path, st);
		seteuid (euid);
		return (rc);
	}

	if (host) {     /* DPM managed file */
		if (Csec_uid) {
			Cns_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				Cns_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		rc = rfio_stat64 (path, st);
		seteuid (euid);
		return (rc);
	}

	/* local file */

	delay_signaling();
	seteuid (0);
	rc = stat64 (filename, st);
	seteuid (euid);
	enable_signaling();
	if (rc < 0)
		return (-1);
	if ((st->st_mode & S_IFMT) != S_IFREG ||
	    (st->st_uid != dpm_uid &&
	    st->st_gid != dpm_gid)) {   /* not a file managed by DPM */
		return (stat64 (filename, st));
	}

	/* DPM managed file */

	strcpy (sfn1, data_interface);
	p = sfn1 + strlen (sfn1);
	*p++ = ':';
	strcpy (p, filename);
	if (Csec_uid) {
		Cns_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
		if (voms_ctx.voname && voms_ctx.fqan)
			Cns_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
					voms_ctx.nbfqan);
	}
	seteuid (0);
	rc = Cns_statr (sfn1, &statbuf);
	seteuid (euid);
	if (rc < 0)
		return (-1);
	st->st_uid = statbuf.uid;
	st->st_gid = statbuf.gid;
	st->st_mode = statbuf.filemode;
	return (0);
}

int dpm_handle_statc(char *path, struct Cns_filestatg *statbuf, char *id)
{
	uid_t euid;
	char *filename;
	char *host;
	char *p;
	int rc;
	char sfn1[CA_MAXSFNLEN+1];
	struct stat64 st;

	if ((rc = rfio_parse (path, &host, &filename)) < 0)
		return (-1);
	euid = geteuid();

	if (rc) {       /* remote file */
		seteuid (0);
		rc = rfio_stat64 (path, &st);
		seteuid (euid);
		if (rc < 0)
			return (rc);
		if (Csec_uid) {
			Cns_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				Cns_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		rc = Cns_statr (path, statbuf);
		seteuid (euid);
		if (rc < 0 && serrno == 2)	/* file exists on disk, but not in DPNS */
			rc = 1;	/* probably not a file managed by DPM */
		return (rc);
	}
															     
	if (host) {     /* DPM managed file */
		if (Csec_uid) {
			Cns_client_setAuthorizationId (Csec_uid, Csec_gid, "GSI", id);
			if (voms_ctx.voname && voms_ctx.fqan)
				Cns_client_setVOMS_data (voms_ctx.voname, voms_ctx.fqan,
						voms_ctx.nbfqan);
		}
		seteuid (0);
		rc = Cns_statg (path, NULL, statbuf);
		seteuid (euid);
		return (rc);
	}

	/* local file */

	seteuid (0);
	rc = stat64 (filename, &st);
	seteuid (euid);
	if (rc < 0)
		return (-1);
	if ((st.st_mode & S_IFMT) != S_IFREG ||
	    (st.st_uid != dpm_uid &&
	    st.st_gid != dpm_gid)) {   /* not a file managed by DPM */
		return (1);
	}

	/* DPM managed file */

	strcpy (sfn1, data_interface);
	p = sfn1 + strlen (sfn1);
	*p++ = ':';
	strcpy (p, filename);
	seteuid (0);
	rc = Cns_statr (sfn1, statbuf);
	seteuid (euid);
	return (rc);
}
